/*******************************************
 *
 * Interbench - Interactivity benchmark
 *
 * Author:  Con Kolivas <kernel@kolivas.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *******************************************/

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sched.h>
#include <time.h>
#include <errno.h>
#include <semaphore.h>
#include <pthread.h>
#include <math.h>
#include <sys/utsname.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>

#define MAX_UNAME_LENGTH	100
#define MAX_LOG_LENGTH		((MAX_UNAME_LENGTH) + 4)
#define MIN_BLK_SIZE		1024
#define DEFAULT_RESERVE		64
#define MB			(1024 * 1024)	/* 2^20 bytes */
#define KB			1024
#define MAX_MEM_IN_MB		(1024 * 64)	/* 64 GB */

struct user_data {
	unsigned long loops_per_ms;
	unsigned long ram, swap;
	int duration;
	int do_rt;
	int bench_nice;
	int load_nice;
	unsigned long cpu_load;
	char logfilename[MAX_LOG_LENGTH];
	int log;
	char unamer[MAX_UNAME_LENGTH];
	char datestamp[13];
	FILE *logfile;
} ud = {
	.duration = 30,
	.cpu_load = 4,
	.log = 1,
};

struct sems {
	sem_t ready;
	sem_t start;
	sem_t stop;
	sem_t complete;
	sem_t stopchild;
};

/* Pipes main to/from load and bench processes */
static int m2l[2], l2m[2], m2b[2], b2m[2];

#define MAX_SAMPLES	10000	/* Limit the number of samples */
#define THREADS		11	/* The total number of different loads */

struct data_table {
	unsigned long total_latency;
	unsigned long max_latency;
	unsigned long nr_samples;
	unsigned long deadlines_met;
	unsigned long missed_latency;
	unsigned long missed_deadlines;
};

struct tk_thread {
	struct sems sem;
	unsigned long sleep_interval;
	unsigned long slept_interval;
};

struct thread {
	void (*name)(struct thread *);
	char *label;
	int bench;		/* This thread is suitable for benchmarking */
	int rtbench;		/* Suitable for real time benchmarking */
	int load;		/* Suitable as a background load */
	int nodeadlines;	/* Deadlines_met are meaningless for this load */
	unsigned long decasecond_deadlines;	/* Expected deadlines / 10s */
	pthread_t pthread;
	pthread_t tk_pthread;
	struct sems sem;
	struct data_table benchmarks[THREADS + 1], *dt;
	unsigned long sample[MAX_SAMPLES];
	struct tk_thread tkthread;
	long threadno;
};

/*
 * To add another load or a benchmark you need to increment the value of
 * THREADS, add a function prototype for your function and add an entry to
 * the threadlist. To specify whether the function is a benchmark or a load
 * set the benchmark and/or load flag as appropriate. The basic requirements
 * of a new load can be seen by using emulate_none as a template.
 */

void emulate_none(struct thread *th);
void emulate_audio(struct thread *th);
void emulate_video(struct thread *th);
void emulate_x(struct thread *th);
void emulate_game(struct thread *th);
void emulate_burn(struct thread *th);
void emulate_write(struct thread *th);
void emulate_read(struct thread *th);
void emulate_ring(struct thread *th);
void emulate_compile(struct thread *th);
void emulate_memload(struct thread *th);

struct thread threadlist[THREADS] = {
	{.label = "None", .name = emulate_none, .load = 1},
	{.label = "Audio", .name = emulate_audio, .bench = 1, .rtbench = 1},
	{.label = "Video", .name = emulate_video, .bench = 1, .rtbench = 1, .load = 1},
	{.label = "X", .name = emulate_x, .bench = 1, .load = 1},
	{.label = "Gaming", .name = emulate_game, .bench = 0, .nodeadlines = 1},	/* No useful output yet */
	{.label = "Burn", .name = emulate_burn, .load = 1},
	{.label = "Write", .name = emulate_write, .load = 1},
	{.label = "Read", .name = emulate_read, .load = 1},
	{.label = "Ring", .name = emulate_ring, .load = 0},	/* No useful data from this */
	{.label = "Compile", .name = emulate_compile, .load = 1},
	{.label = "Memload", .name = emulate_memload, .load = 1},
};

void init_sem(sem_t *sem);
void init_all_sems(struct sems *s);
void initialise_thread(int i);
void start_thread(struct thread *th);
void stop_thread(struct thread *th);

void terminal_error(char *name)
{
	perror(name);
	exit (1);
}

void terminal_fileopen_error(FILE *fp, char *name)
{
	perror(name);
	if (fclose(fp) == -1)
		terminal_error("fclose");
	exit (1);
}

unsigned long long get_nsecs(struct timespec *myts)
{
	if (clock_gettime(CLOCK_MONOTONIC, myts))
		terminal_error("clock_gettime");
	return (myts->tv_sec * 1000000000 + myts->tv_nsec );
}

unsigned long get_usecs(struct timespec *myts)
{
	if (clock_gettime(CLOCK_MONOTONIC, myts))
		terminal_error("clock_gettime");
	return (myts->tv_sec * 1000000 + myts->tv_nsec / 1000 );
}

void set_fifo(int prio)
{
	struct sched_param sp;
	memset(&sp, 0, sizeof(sp));
	sp.sched_priority = prio;
	if (sched_setscheduler(0, SCHED_FIFO, &sp) == -1) {
		if (errno != EPERM)
			terminal_error("sched_setscheduler");
		fprintf(stderr, "could not set RT scheduling\n");
	}
}

void set_thread_fifo(pthread_t pthread, int prio)
{
	struct sched_param sp;
	memset(&sp, 0, sizeof(sp));
	sp.sched_priority = prio;
	if (pthread_setschedparam(pthread, SCHED_FIFO, &sp) == -1)
		terminal_error("pthread_setschedparam");
}

void set_normal(void)
{
	struct sched_param sp;
	memset(&sp, 0, sizeof(sp));
	sp.sched_priority = 0;
	if (sched_setscheduler(0, SCHED_OTHER, &sp) == -1) {
		fprintf(stderr, "could not unset RT scheduling!\n");
	}
}

void set_nice(int prio)
{
	if (setpriority(PRIO_PROCESS, 0, prio) == -1)
		terminal_error("setpriority");
}

int test_fifo(void)
{
	struct sched_param sp;
	memset(&sp, 0, sizeof(sp));
	sp.sched_priority = 99;
	if (sched_setscheduler(0, SCHED_FIFO, &sp) == -1) {
		if (errno != EPERM)
			terminal_error("sched_setscheduler");
		goto out_fail;
	}
	if (sched_getscheduler(0) != SCHED_FIFO)
		goto out_fail;
	set_normal();
	return 1;
out_fail:
	set_normal();
	return 0;
}

void set_thread_normal(pthread_t pthread)
{
	struct sched_param sp;
	memset(&sp, 0, sizeof(sp));
	sp.sched_priority = 0;
	if (pthread_setschedparam(pthread, SCHED_OTHER, &sp) == -1)
		terminal_error("pthread_setschedparam");
}

unsigned long compute_allocable_mem(void)
{
	unsigned long total = ud.ram + ud.swap;
	unsigned long usage = ud.ram * 110 / 100 ;

	/* Leave at least DEFAULT_RESERVE free space and check for maths overflow. */
	if (total - DEFAULT_RESERVE < usage)
		usage = total - DEFAULT_RESERVE;
	usage /= 1024;	/* to megabytes */
	if (usage > 2930)
		usage = 2930;
	return usage;
}

void burn_loops(unsigned long loops)
{
	unsigned long i;

	/* 
	 * We need some magic here to prevent the compiler from optimising
	 * this loop away. Otherwise trying to emulate a fixed cpu load
	 * with this loop will not work.
	 */
	for (i = 0 ; i < loops ; i++)
	     asm volatile("" : : : "memory");
}

/* Use this many usecs of cpu time */
void burn_usecs(unsigned long usecs)
{
	unsigned long ms_loops;

	ms_loops = ud.loops_per_ms / 1000 * usecs;
	burn_loops(ms_loops);
}

void microsleep(unsigned long long usecs)
{
	struct timespec req;

	req.tv_sec = usecs / 1000000;
	req.tv_nsec = (usecs - (req.tv_sec * 1000000)) * 1000;
	nanosleep(&req, NULL);
}

unsigned long periodic_schedule(struct thread *th, unsigned long run_usecs,
	unsigned long interval_usecs, unsigned long deadline)
{
	unsigned long current_time, latency;
	struct tk_thread *tk;
	struct data_table *tb;
	struct timespec myts;

	latency = 0;
	tb = th->dt;
	tk = &th->tkthread;

	current_time = get_usecs(&myts);
	/* First sample latency is ignored */
	if (!tb->nr_samples)
		current_time = deadline - 1;
	if (current_time > deadline)
		latency = current_time - deadline;
	if (tk->slept_interval < latency)
		latency -= tk->slept_interval;
	if (tb->nr_samples < MAX_SAMPLES)
		th->sample[tb->nr_samples] = latency;
	tb->total_latency += latency;
	latency += tb->missed_latency;
	if (latency > tb->max_latency)
		tb->max_latency = latency;

	burn_usecs(run_usecs);
	current_time = get_usecs(&myts);
	/*
	 * If we meet the deadline we move the deadline forward, otherwise
	 * we consider it a missed deadline and dropped frame etc.
	 */
	deadline += interval_usecs;
	if (deadline > current_time) {
		tb->deadlines_met++;
		tb->total_latency += tb->missed_latency;
		tb->missed_latency = 0;
	} else if (interval_usecs) {
		unsigned long intervals = (current_time - deadline) /
			interval_usecs + 1;

		tb->missed_deadlines += intervals;
		tb->missed_latency += intervals * interval_usecs;
		deadline += intervals * interval_usecs;
	} else
		goto out_nosleep;
	tk->sleep_interval = deadline - current_time;
	sem_post(&tk->sem.start);
	sem_wait(&tk->sem.complete);
out_nosleep:
	return deadline;
}

void initialise_thread_data(struct thread *th)
{
	struct data_table *tb = th->dt;

	tb->max_latency =
	tb->total_latency =
	tb->deadlines_met =
	tb->missed_latency =
	tb->missed_deadlines =
	tb->nr_samples = 0;
}

void create_pthread(pthread_t  * thread, pthread_attr_t * attr,
	void * (*start_routine)(void *), void *arg)
{
	if (pthread_create(thread, attr, start_routine, arg))
		terminal_error("pthread_create");
}

void join_pthread(pthread_t th, void **thread_return)
{
	if (pthread_join(th, thread_return))
		terminal_error("pthread_join");
}

void emulate_none(struct thread *th)
{
	sem_t *s = &th->sem.stop;
	sem_wait(s);
}

#define audio_interval	(50000)
#define audio_run	(audio_interval / 20)
/* We emulate audio by using 5% cpu and waking every 50ms */
void emulate_audio(struct thread *th)
{
	unsigned long deadline;
	sem_t *s = &th->sem.stop;
	struct timespec myts;

	th->decasecond_deadlines = 1000000 / audio_interval * 10;
	deadline = get_usecs(&myts);

	while (1) {
		deadline = periodic_schedule(th, audio_run, audio_interval,
			deadline);
		if (!sem_trywait(s))
			return;
		th->dt->nr_samples++;
	}
}

/* We emulate video by using 40% cpu and waking for 60fps */
#define video_interval	(1000000 / 60)
#define video_run	(video_interval * 40 / 100)
void emulate_video(struct thread *th)
{
	unsigned long deadline;
	sem_t *s = &th->sem.stop;
	struct timespec myts;

	th->decasecond_deadlines = 1000000 / video_interval * 10;
	deadline = get_usecs(&myts);

	while (1) {
		deadline = periodic_schedule(th, video_run, video_interval,
			deadline);
		if (!sem_trywait(s))
			return;
		th->dt->nr_samples++;
	}
}

/*
 * We emulate X by running for a variable percentage of cpu from 0-100% 
 * in 1ms chunks.
 */
void emulate_x(struct thread *th)
{
	unsigned long deadline;
	sem_t *s = &th->sem.stop;
	struct timespec myts;

	th->decasecond_deadlines = 100;
	deadline = get_usecs(&myts);

	while (1) {
		int i, j;
		for (i = 0 ; i <= 100 ; i++) {
			j = 100 - i;
			deadline = periodic_schedule(th, i * 1000, j * 1000,
				deadline);
			deadline += i * 1000;
			if (!sem_trywait(s))
				return;
			th->dt->nr_samples++;
		}
	}
}

/* 
 * We emulate gaming by using 100% cpu and seeing how many frames (jobs
 * completed) we can do in that time. Deadlines are meaningless with 
 * unlocked frame rates
 */
#define game_interval	(10000)
#define game_run	(game_interval)
void emulate_game(struct thread *th)
{
	unsigned long deadline;
	sem_t *s = &th->sem.stop;
	struct timespec myts;
	th->decasecond_deadlines = 1000000 / game_interval * 10;

	deadline = get_usecs(&myts);

	while (1) {
		deadline = periodic_schedule(th, game_run, game_interval,
			deadline);
		if (!sem_trywait(s))
			return;
		th->dt->nr_samples++;
	}
}

void *burn_thread(void * t)
{
	struct thread *th;
	sem_t *s;
	long i = (long)t;

	th = &threadlist[i];
	s = &th->sem.stopchild;

	while (1) {
		burn_loops(ud.loops_per_ms);
		if (!sem_trywait(s)) {
			sem_post(s);
			break;
		}
	}
	return NULL;
}

/* Have ud.cpu_load threads burn cpu continuously */
void emulate_burn(struct thread *th)
{
	sem_t *s = &th->sem.stop;
	unsigned long i;
	long t;
	pthread_t burnthreads[ud.cpu_load];

	t = th->threadno;
	for (i = 0 ; i < ud.cpu_load ; i++)
		create_pthread(&burnthreads[i], NULL, burn_thread,
			(void*)(long) t);
	sem_wait(s);
	sem_post(&th->sem.stopchild);
	for (i = 0 ; i < ud.cpu_load ; i++)
		join_pthread(burnthreads[i], NULL);
}

/* Write a file the size of ram continuously */
void emulate_write(struct thread *th)
{
	sem_t *s = &th->sem.stop;
	FILE *fp;
	char *name = "interbench.write";
	void *buf = NULL;
	struct stat statbuf;
	unsigned long mem;

	if (!(fp = fopen(name, "w")))
		terminal_error("fopen");
	if (stat(name, &statbuf) == -1)
		terminal_fileopen_error(fp, "stat");
	if (statbuf.st_blksize < MIN_BLK_SIZE)
		statbuf.st_blksize = MIN_BLK_SIZE;
	mem = ud.ram / (statbuf.st_blksize / 1024);	/* kilobytes to blocks */
	if (!(buf = calloc(1, statbuf.st_blksize)))
		terminal_fileopen_error(fp, "calloc");
	if (fclose(fp) == -1)
		terminal_error("fclose");

	while (1) {
		unsigned int i;

		if (!(fp = fopen(name, "w")))
			terminal_error("fopen");
		if (stat(name, &statbuf) == -1)
			terminal_fileopen_error(fp, "stat");
		for (i = 0 ; i < mem; i++) {
			if (fwrite(buf, statbuf.st_blksize, 1, fp) != 1)
				terminal_fileopen_error(fp, "fwrite");
			if (!sem_trywait(s))
				goto out;
		}
		if (fclose(fp) == -1)
			terminal_error("fclose");
	}

out:
	if (fclose(fp) == -1)
		terminal_error("fclose");
	if (remove(name) == -1)
		terminal_error("remove");
	fflush(NULL);
	sync();
}

/* Read a file the size of ram continuously */
void emulate_read(struct thread *th)
{
	sem_t *s = &th->sem.stop;
	char *name = "interbench.read";
	void *buf = NULL;
	struct stat statbuf;
	unsigned long bsize;
	int tmp;

	if ((tmp = open(name, O_RDONLY)) == -1)
		terminal_error("open");
	if (stat(name, &statbuf) == -1) 
		terminal_error("stat");
	bsize = statbuf.st_blksize;
	if (!(buf = malloc(bsize)))
		terminal_error("malloc");

	while (1) {
		int rd;

		/* 
		 * We have to read the whole file before quitting the load
		 * to prevent the data being cached for the next read. This
		 * is also the reason the file is the size of physical ram.
		 */
		while ((rd = read(tmp , buf, bsize)) > 0);
		if(!sem_trywait(s))
			return;
		if (lseek(tmp, (off_t)0, SEEK_SET) == -1)
			terminal_error("lseek");
	}
}

#define RINGTHREADS	4

struct thread ringthreads[RINGTHREADS];

void *ring_thread(void * t)
{
	struct thread *th;
	struct sems *s;
	int i, post_to;

	i = (long)t;
	th = &ringthreads[i];
	s = &th->sem;
	post_to = i + 1;
	if (post_to == RINGTHREADS)
		post_to = 0;
	if (i == 0)
		sem_post(&s->ready);

	while (1) {
		sem_wait(&s->start);
		sem_post(&ringthreads[post_to].sem.start);
		if (!sem_trywait(&s->stop))
			goto out;
	}
out:	
	sem_post(&ringthreads[post_to].sem.start);
	sem_post(&s->complete);
	return NULL;
}

/* Create a ring of 4 processes that wake each other up in a circle */
void emulate_ring(struct thread *th)
{
	sem_t *s = &th->sem.stop;
	int i;

	for (i = 0 ; i < RINGTHREADS ; i++) {
		init_all_sems(&ringthreads[i].sem);
		create_pthread(&ringthreads[i].pthread, NULL, 
			ring_thread, (void*)(long) i);
	}

	sem_wait(&ringthreads[0].sem.ready);
	sem_post(&ringthreads[0].sem.start);
	sem_wait(s);
	for (i = 0 ; i < RINGTHREADS ; i++)
		sem_post(&ringthreads[i].sem.stop);
	for (i = 0 ; i < RINGTHREADS ; i++) {
		sem_wait(&ringthreads[i].sem.complete);
		join_pthread(ringthreads[i].pthread, NULL);
	}
}

/* We emulate a compile by running burn, write and read threads simultaneously */
void emulate_compile(struct thread *th)
{
	sem_t *s = &th->sem.stop;
	unsigned long i, threads[3];

	for (i = 0 ; i < THREADS ; i++) {
		if (threadlist[i].label == "Burn")
			threads[0] = i;
		if (threadlist[i].label == "Write")
			threads[1] = i;
		if (threadlist[i].label == "Read")
			threads[2] = i;
	}
	for (i = 0 ; i < 3 ; i++) {
		if (!threads[i]) {
			fprintf(stderr, "Can't find all threads for compile load\n");
			exit(1);
		}
	}
	for (i = 0 ; i < 3 ; i++) {
		initialise_thread(threads[i]);
		start_thread(&threadlist[threads[i]]);
	}
	sem_wait(s);
	for (i = 0 ; i < 3 ; i++)
		stop_thread(&threadlist[threads[i]]);
}

int *grab_and_touch (char *block[], int i)
{
	block[i] = (char *) malloc(MB);
	if (!block[i])
		return NULL;
	return (memset(block[i], 1, MB));
}

/* We emulate a memory load by allocating and torturing 110% of available ram */
void emulate_memload(struct thread *th)
{
	sem_t *s = &th->sem.stop;
	unsigned long touchable_mem, i;
	char *mem_block[MAX_MEM_IN_MB];
	void *success;

	touchable_mem = compute_allocable_mem();
	/* loop until we're killed, frobbing memory in various perverted ways */
	while (1) {
		for (i = 0;  i < touchable_mem; i++) {
			success = grab_and_touch(mem_block, i);
			if (!sem_trywait(s))
				goto out_freemem;
			if (!success) {
				touchable_mem = i-1;
				break;
			}
		}
		for (i = 0;  i < touchable_mem; i++) {
			memcpy(mem_block[i], mem_block[(i + touchable_mem / 2) %
				touchable_mem], MB);
			if (!sem_trywait(s))
				goto out_freemem;
		}
		for (i = 0; i < touchable_mem; i++) {
			free(mem_block[i]);
		}	
		if (!sem_trywait(s))
			goto out;
	}
out_freemem:
	for (i = 0; i < touchable_mem; i++)
		free(mem_block[i]);
out:
	return;
}

void *timekeeping_thread(void * t)
{
	struct thread *th;
	struct tk_thread *tk;
	struct sems *s;
	struct timespec myts;
	long i = (long)t;

	th = &threadlist[i];
	tk = &th->tkthread;
	s = &th->tkthread.sem;
	/*
	 * If this timekeeping thread is not that of a benchmarked thread we
	 * run lower priority than a real time benchmarked thread would run.
	 */
	if (th->dt == &th->benchmarks[THREADS])	/* Not being benchmarked */
		set_fifo(94);
	else
		set_fifo(96);
	/* These values must be changed at the appropriate places or race */
	tk->sleep_interval = 0;
	tk->slept_interval = 0;
	sem_post(&s->ready);

	while (1) {
		unsigned long now;

		if (!sem_trywait(&s->stop))
			goto out;
		sem_wait(&s->start);
		now = get_usecs(&myts);
		if (!sem_trywait(&s->stop))
			goto out;
		if (tk->sleep_interval)
			microsleep(tk->sleep_interval);
		tk->slept_interval = get_usecs(&myts) - now -
			tk->sleep_interval;
		tk->sleep_interval = 0;
		sem_post(&s->complete);
	}
out:
	return NULL;
}

/*
 * All the sleep functions such as nanosleep can only guarantee that they
 * sleep for _at least_ the time requested. We work around this by having
 * a high priority real time thread that accounts for the extra time slept
 * in nanosleep. This allows wakeup latency of the tested thread to be
 * accurate and reflect true scheduling delays.
 */
void *emulation_thread(void * t)
{
	struct thread *th;
	struct tk_thread *tk;
	struct sems *s, *tks;
	long i = (long)t;

	th = &threadlist[i];
	tk = &th->tkthread;
	s = &th->sem;
	tks = &tk->sem;
	init_all_sems(tks);

	/* Start the timekeeping thread */
	create_pthread(&th->tk_pthread, NULL, timekeeping_thread,
		(void*)(long) i);
	/* Wait for timekeeping thread to be ready */
	sem_wait(&tks->ready);

	/* Tell main we're ready to start*/
	sem_post(&s->ready);

	/* Wait for signal from main to start thread */
	sem_wait(&s->start);

	/* Start the actual function being benched/or running as load */
	th->name(th);

	/* Stop the timekeeping thread */
	sem_post(&tks->stop);
	sem_post(&tks->start);
	join_pthread(th->tk_pthread, NULL);

	/* Tell main we've finished */
	sem_post(&s->complete);
	return NULL;
}

/*
 * In an unoptimised loop we try to benchmark how many meaningless loops
 * per second we can perform on this hardware to fairly accurately
 * reproduce certain percentage cpu usage
 */
void calibrate_loop(void)
{
	unsigned long long start_time, loops_per_msec, run_time = 0;
	unsigned long loops;
	struct timespec myts;

	loops_per_msec = 100000;
redo:
	while (run_time / 1000 != 1000) {
		loops = loops_per_msec;
		start_time = get_nsecs(&myts);
		burn_loops(loops);
		run_time = get_nsecs(&myts) - start_time;
		loops_per_msec = (1000000 * loops_per_msec / run_time ? :
			loops_per_msec);
	}

	/* Rechecking after a pause increases reproducibility */
	sleep(1);
	loops = loops_per_msec;
	start_time = get_nsecs(&myts);
	burn_loops(loops);
	run_time = get_nsecs(&myts) - start_time;
	if (run_time / 1000 != 1000)
		goto redo;

	ud.loops_per_ms = loops_per_msec;
}

void log_output(const char *format, ...) __attribute__ ((format(printf, 1, 2)));

/* Output to console +/- logfile */
void log_output(const char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	if (vprintf(format, ap) == -1)
		terminal_error("vprintf");
	va_end(ap);
	if (ud.log) {
		va_start(ap, format);
		if (vfprintf(ud.logfile, format, ap) == -1)
			terminal_error("vpfrintf");
		va_end(ap);
	}
	fflush(NULL);
}

/* Calculate statistics and output them */
void show_latencies(struct thread *th)
{
	struct data_table *tbi = &th->benchmarks[0],
		*tbj = th->dt;
	double average_latency, max_latency, deadlines_met, samples_met, sd;
	unsigned long nr_samples, noload_samples, expected_deadlines, s;
	long double variance = 0;

	expected_deadlines = th->decasecond_deadlines * (ud.duration / 10);
	if (tbi->nr_samples > 1)
		noload_samples = tbi->nr_samples - 1;
	else
		noload_samples = 1;
	if (noload_samples >= MAX_SAMPLES)
		noload_samples = MAX_SAMPLES - 1;
	nr_samples = tbj->nr_samples ? : 1;
	if (nr_samples >= MAX_SAMPLES)
		nr_samples = MAX_SAMPLES - 1;
	average_latency = tbj->total_latency / nr_samples;
	for (s = 0 ; s < nr_samples; s++) {
		long diff = average_latency - th->sample[s];
		variance += diff * diff;
	}
	variance /= (nr_samples - 1) ? : 1;
	sd = sqrtl(variance);

	/* 
	 * Landing on the boundary of a deadline can make loaded runs appear
	 * to do more work than unloaded due to tiny duration differences.
	 */
	if (nr_samples > noload_samples)
		nr_samples = noload_samples;
	samples_met = nr_samples / noload_samples * 100;
	if (samples_met > 100)
		samples_met = 100;
	max_latency = tbj->max_latency;
	/* When benchmarking rt we represent the data in us */
	if (!ud.do_rt) {
		average_latency /= 1000;
		sd /= 1000;
		max_latency /= 1000;
	}
	deadlines_met = tbj->deadlines_met / expected_deadlines * 100;
	if (deadlines_met > 100)
		deadlines_met = 100;

	if (average_latency >= 100)
		log_output("%7.0f +/- ", average_latency);
	else
		log_output("%7.3g +/- ", average_latency);
	if (sd >= 100)
		log_output("%-9.0f", sd);
	else
		log_output("%-9.3g", sd);
	if (!th->nodeadlines) {
		if (max_latency >= 100)
			log_output("%7.0f\t", max_latency);
		else
			log_output("%7.3g\t", max_latency);
	}
	log_output("\t%4.3g", samples_met);
	if (!th->nodeadlines)
		log_output("\t%11.3g", deadlines_met);
	log_output("\n");
}

void create_read_file(void)
{
	unsigned int i;
	FILE *fp;
	char *name = "interbench.read";
	void *buf = NULL;
	struct stat statbuf;
	unsigned long mem, bsize;
	int tmp;

	if ((tmp = open(name, O_RDONLY)) == -1) {
		if (errno != ENOENT)
			terminal_error("open");
		goto write;
	}
	if (stat(name, &statbuf) == -1)
		terminal_error("stat");
	if (statbuf.st_blksize < MIN_BLK_SIZE)
		statbuf.st_blksize = MIN_BLK_SIZE;
	bsize = statbuf.st_blksize;
	if (statbuf.st_size / 1024 / bsize == ud.ram / bsize)
		return;
	if (remove(name) == -1)
		terminal_error("remove");
write:
	fprintf(stderr,"Creating file for read load...\n");
	if (!(fp = fopen(name, "w")))
		terminal_error("fopen");
	if (stat(name, &statbuf) == -1)
		terminal_fileopen_error(fp, "stat");
	if (statbuf.st_blksize < MIN_BLK_SIZE)
		statbuf.st_blksize = MIN_BLK_SIZE;
	bsize = statbuf.st_blksize;
	if (!(buf = calloc(1, bsize)))
		terminal_fileopen_error(fp, "calloc");
	mem = ud.ram / (bsize / 1024);	/* kilobytes to blocks */

	for (i = 0 ; i < mem; i++) {
		if (fwrite(buf, bsize, 1, fp) != 1)
			terminal_fileopen_error(fp, "fwrite");
	}
	if (fclose(fp) == -1)
		terminal_error("fclose");
	fflush(NULL);
	sync();
}

void get_ram(void)
{
	FILE *meminfo;

	if(!(meminfo = fopen("/proc/meminfo", "r")))
		terminal_error("fopen");
	fscanf(meminfo, "MemTotal: %lu kB", &ud.ram);
	fscanf(meminfo, "SwapTotal: %lu kB", &ud.swap);
	if (fclose(meminfo) == -1)
		terminal_error("fclose");
}

void get_logfilename(void)
{
	struct tm *mytm;
	struct utsname buf;
	time_t t;
	int year, month, day, hours, minutes;

	time(&t);
	if (uname(&buf) == -1)
		terminal_error("uname");
	if (!(mytm = localtime(&t)))
		terminal_error("localtime");
	year = mytm->tm_year + 1900;
	month = mytm->tm_mon + 1;
	day = mytm->tm_mday;
	hours = mytm->tm_hour;
	minutes = mytm->tm_min;
	strncpy(ud.unamer, buf.release, MAX_UNAME_LENGTH);

	sprintf(ud.datestamp, "%2d%02d%02d%02d%02d",
		year, month, day, hours, minutes);
	snprintf(ud.logfilename, MAX_LOG_LENGTH, "%s.log", ud.unamer);
}

void start_thread(struct thread *th)
{
	sem_post(&th->sem.start);
}

void stop_thread(struct thread *th)
{
	sem_post(&th->sem.stop);
	sem_wait(&th->sem.complete);

	/* Kill the thread */
	join_pthread(th->pthread, NULL);
}

void init_sem(sem_t *sem)
{
	if (sem_init(sem, 1, 0))
		terminal_error("sem_init");
}

void init_all_sems(struct sems *s)
{
	/* Initialise the semaphores */
	init_sem(&s->ready);
	init_sem(&s->start);
	init_sem(&s->stop);
	init_sem(&s->complete);
	init_sem(&s->stopchild);
}

void initialise_thread(int i)
{
	struct thread *th = &threadlist[i];

	init_all_sems(&th->sem);
	/* Create the threads. Yes, the (long) cast is fugly but it's safe*/
	create_pthread(&th->pthread, NULL, emulation_thread, (void*)(long)i);

	sem_wait(&th->sem.ready);
	/* 
	 * The data in benchmarks[THREADS] is unused, so it will be set to
	 * the actual load being run only if this thread is being benchmarked
	 */
	th->dt = &th->benchmarks[THREADS];
	initialise_thread_data(th);
	
}

/* A pseudo-semaphore for processes using a pipe */
void wait_on(int *pype)
{
	int retval, buf;

	do {
		retval = read(*pype, &buf, sizeof(buf));
	} while (retval == -1 && errno == EINTR);
}

void wakeup_with(int *pype)
{
	int retval, buf = 1;

	do {
		retval = write(*pype, &buf, sizeof(buf));
	} while (retval == -1 && errno == EINTR);
}

void run_loadchild(int j)
{
	struct thread *thj;
	thj = &threadlist[j];

	set_nice(ud.load_nice);
	initialise_thread(j);

	/* Tell main we're ready */
	wakeup_with(&l2m[1]);

	/* Main tells us we're ready */
	wait_on(&m2l[0]);
	start_thread(thj);

	/* Main tells us to stop */
	wait_on(&m2l[0]);
	stop_thread(thj);

	/* Tell main we've finished */
	wakeup_with(&l2m[1]);
	exit(0);
}

void run_benchchild(int i, int j)
{
	struct thread *thi;

	thi = &threadlist[i];

	set_nice(ud.bench_nice);
	initialise_thread(i);
	/* Point the data table to the appropriate load being tested */
	thi->dt = &thi->benchmarks[j];
	initialise_thread_data(thi);
	if (ud.do_rt)
		set_thread_fifo(thi->pthread, 95);
	
	/* Tell main we're ready */
	wakeup_with(&b2m[1]);

	/* Main tells us we're ready */
	wait_on(&m2b[0]);
	start_thread(thi);
	
	/* Main tells us to stop */
	wait_on(&m2b[0]);
	stop_thread(thi);
	if (ud.do_rt)
		set_thread_normal(thi->pthread);
	show_latencies(thi);

	/* Tell main we've finished */
	wakeup_with(&b2m[1]);
	exit(0);
}

void bench(int i, int j)
{
	pid_t bench_pid, load_pid;

	if ((load_pid = fork()) == -1)
		terminal_error("fork");
	if (!load_pid) {
		if (close(m2l[1]) == -1 || close(l2m[0]) == -1)
			terminal_error("close");
		run_loadchild(j);
	}
	if ((bench_pid = fork()) == -1)
		terminal_error("fork");
	if (!bench_pid){
		if (close(m2b[1]) == -1 || close(b2m[0]) == -1)
			terminal_error("close");
		run_benchchild(i, j);
	}

	/* Wait for load and bench processes to be ready */
	wait_on(&l2m[0]);
	wait_on(&b2m[0]);

	/* Wakeup the load process */
	wakeup_with(&m2l[1]);
	sleep(1);

	/* We want to be higher priority than everything to signal the stop */
	set_fifo(99);

	/* After a small delay, wake up the benched process */
	wakeup_with(&m2b[1]);
	microsleep(ud.duration * 1000000);

	/* Tell the benched process to stop its threads and output results */
	wakeup_with(&m2b[1]);
	/* Tell the load process to stop its threads */
	wakeup_with(&m2l[1]);

	/* Return to SCHED_NORMAL */
	set_normal();

	/* Wait for load and bench processes to terminate */
	wait_on(&l2m[0]);
	wait_on(&b2m[0]);
}

void init_pipe(int *pype)
{
	if (pipe(pype) == -1)
		terminal_error("pipe");
}

void init_pipes(void)
{
	init_pipe(m2l);
	init_pipe(l2m);
	init_pipe(m2b);
	init_pipe(b2m);
}

void usage(void)
{
	/* Affinity commented out till working on all architectures */
	fprintf(stderr, "interbench v0.21 by Con Kolivas\n");
	//fprintf(stderr, "interbench [-l <int>] [-L <int>] [-t <int] [-u] [-b] [-c] [-h] [-n] [-r] \n");
	fprintf(stderr, "interbench [-l <int>] [-L <int>] [-t <int] [-B <int>] [-N <int>] [-b] [-c] [-h] [-n] [-r] \n");
	fprintf(stderr, " -l\tUse <int> loops per sec (default: use saved benchmark)\n");
	fprintf(stderr, " -L\tUse cpu load of <int> with burn load (default: 4)\n");
	fprintf(stderr, " -t\tSeconds to run each benchmark (default: 30)\n");
	fprintf(stderr, " -B\tNice the benchmarked thread to <int> (default: 0)\n");
	fprintf(stderr, " -N\tNice the load thread to <int> (default: 0)\n");
	//fprintf(stderr, " -u\tImitate uniprocessor\n");
	fprintf(stderr, " -b\tBenchmark loops_per_ms even if it is already known\n");
	fprintf(stderr, " -c\tOutput to console only (default: use console and logfile)\n");
	fprintf(stderr, " -r\tPerform real time scheduling benchmarks (default: non-rt)\n");
	fprintf(stderr, " -h\tShow this help\n");
}

int main(int argc, char **argv)
{
	int q, i, j, affinity, benchmark = 0;
	FILE *fp;
	/* 
	 * This file stores the loops_per_ms to be reused in a filename that
	 * can't be confused
	 */
	char *fname = "interbench.loops_per_ms";

	while ((q = getopt(argc, argv, "hl:L:B:N:ut:bcnr")) != -1) {
		switch (q) {
			case 'h':
				usage();
				return (0);
			case 'l':
				ud.loops_per_ms = atoi(optarg);
				break;
			case 't':
				ud.duration = atoi(optarg);
				break;
			case 'L':
				ud.cpu_load = atoi(optarg);
				break;
			case 'B':
				ud.bench_nice = atoi(optarg);
				break;
			case 'N':
				ud.load_nice = atoi(optarg);
				break;
			case 'u':
				affinity = 1;
				break;
			case 'b':
				benchmark = 1;
				break;
			case 'c':
				ud.log = 0;
				break;
			case 'r':
				ud.do_rt = 1;
				break;
			default:
				usage();
				return (1);
		}
	}
	argc -= optind;
	argv += optind;
	/*FIXME Need to parse more input to take choice of loads */

	if (!test_fifo()) {
		ud.do_rt = 0;
		fprintf(stderr, "Unable to get SCHED_FIFO, results will be unreliable and no real time tests will be conducted\n");
	}
	if (!ud.cpu_load) {
		fprintf(stderr, "Invalid cpu load\n");
		exit (1);
	}

	/*FIXME Affinity commented out till working on all architectures */
#if 0
	if (affinity) {
#ifdef CPU_SET	/* Current glibc expects cpu_set_t */
		cpu_set_t cpumask;

		CPU_ZERO(&cpumask);
		CPU_SET(0, &cpumask);
#else		/* Old glibc expects unsigned long */
		unsigned long cpumask = 1;
#endif
		if (sched_setaffinity(0, sizeof(cpumask), &cpumask) == -1) {
			if (errno != EPERM)
				terminal_error("sched_setaffinity");
			fprintf(stderr, "could not set cpu affinity\n");
		}
	}
#endif

	/* Make benchmark a multiple of 10 seconds for proper range of X loads */
	if (ud.duration % 10)
		ud.duration += 10 - ud.duration % 10;

	if (benchmark)
		ud.loops_per_ms = 0;
	/* 
	 * Try to get loops_per_ms from command line first, file second, and
	 * benchmark if not available.
	 */
	if (!ud.loops_per_ms) {
		if (benchmark)
			goto bench;
		if ((fp = fopen(fname, "r"))) {
			fscanf(fp, "%lu", &ud.loops_per_ms);
			if (fclose(fp) == -1)
				terminal_error("fclose");
			if (ud.loops_per_ms) {
				fprintf(stderr,
					"%lu loops_per_ms read from file interbench.loops_per_ms\n",
					ud.loops_per_ms);
				goto loops_known;
			}
		}
bench:
		fprintf(stderr, "loops_per_ms unknown; benchmarking...\n");

		/*
		 * To get as accurate a loop as possible we time it running
		 * SCHED_FIFO if we can
		 */
		set_fifo(99);
		calibrate_loop();
		set_normal();
	} else
		fprintf(stderr, "loops_per_ms specified from command line\n");

	if (!(fp = fopen(fname, "w"))) {
		fprintf(stderr, "Unable to write to file interbench.loops_per_ms\n");
		goto loops_known;
	}
	fprintf(fp, "%lu", ud.loops_per_ms);
	fprintf(stderr, "%lu loops_per_ms saved to file interbench.loops_per_ms\n",
		ud.loops_per_ms);
	if (fclose(fp) == -1)
		terminal_error("fclose");

loops_known:
	get_ram();
	get_logfilename();
	create_read_file();
	init_pipes();

	if (ud.log && !(ud.logfile = fopen(ud.logfilename, "a"))) {
		fprintf(stderr, "Unable to write to logfile\n");
		ud.log = 0;
	}
	log_output("\n");
	log_output("Using %lu loops per ms, running every load for %d seconds\n",
		ud.loops_per_ms, ud.duration);
	log_output("Benchmarking kernel %s with datestamp %s\n",
		ud.unamer, ud.datestamp);
	log_output("\n");

	for (i = 0 ; i < THREADS ; i++)
		threadlist[i].threadno = i;

	for (i = 0 ; i < THREADS ; i++) {
		struct thread *thi = &threadlist[i];
		int *benchme;

		if (ud.do_rt)
			benchme = &threadlist[i].rtbench;
		else
			benchme = &threadlist[i].bench;

		if (!*benchme)
			continue;

		log_output("--- Benchmarking %s ", threadlist[i].label);
		if (ud.do_rt)
			log_output("real time ");
		else if (ud.bench_nice)
			log_output("nice %d ", ud.bench_nice);
		log_output("in the presence of loads ");
		if (ud.load_nice)
			log_output("nice %d ", ud.load_nice);
		log_output("---\n");

		if (ud.do_rt)
			log_output("\tLatency +/- SD (us)");
		else
			log_output("\tLatency +/- SD (ms)");
		if (!thi->nodeadlines)
			log_output("  Max Latency ");
		log_output("  %% Desired CPU");
		if (!thi->nodeadlines)
			log_output("  %% Deadlines Met");
		log_output("\n");

		for (j = 0 ; j < THREADS ; j++) {
			struct thread *thj = &threadlist[j];

			if (j == i || !threadlist[j].load)
				continue;
			log_output("%s\t", thj->label);
			sync();
			bench(i, j);
		}
		log_output("\n");
	}
	log_output("\n");
	if (ud.log)
		fclose(ud.logfile);

	return 0;
}
